#  tests for comtypes-1.1.9-py36haa95532_1002 (this is a generated file);
print('===== testing package: comtypes-1.1.9-py36haa95532_1002 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
from comtypes.client import CreateObject, CoGetObject, GetActiveObject
from comtypes.client import ShowEvents, PumpEvents, GetEvents
#  --- run_test.py (end) ---

print('===== comtypes-1.1.9-py36haa95532_1002 OK =====');
print("import: 'comtypes'")
import comtypes

